/*
 * Created on Jun 6, 2005
 *
 * Unpublished Work, Copyright, 2005 Northrop Grumman
 */
package gov.grants.apply.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.ConcurrentModificationException;
import java.util.Iterator;

import javax.activation.DataSource;

import org.apache.log4j.Logger;

/**
 * @author Grants.gov WebServices Team
 *
 * A DataSource type object that uses a byte array as the data that it draws from.
 */
public final class ByteArrayDataSource implements DataSource
{
   private static final Logger LOG = Logger.getLogger(ByteArrayDataSource.class);

   private String _contentType = null;
   private String _name = null;
   private byte[] _data = null;
   private ArrayList _inputStreams = null;

   /**
    * Create a data source with the given content type, name, and data.
    * @param contentType The MIME type that describes the data that is in the byte array.
    * @param name The name of the data.
    * @param data The byte array where this data source will draw its data from.
    */
   public ByteArrayDataSource(String contentType, String name, byte[] data)
   {
      _contentType = contentType;
      _name = name;
      _data = data;
   }
   
   /**
    * @return The MIME type that describes the data that is in the byte array.
    */
   public String getContentType()
   {
      return _contentType;
   }

   /**
    * @return A input stream that will allow the client to write into the byte array.
    */
   public InputStream getInputStream() throws IOException
   {
      if (_data == null)
      {
         LOG.warn("Was asked to getInputStream from a null byte array.");
         return null;
      }
      ByteArrayDSInputStream stream = new ByteArrayDSInputStream(new ByteArrayInputStream(_data), this);
      if (_inputStreams == null)
      {
         _inputStreams = new ArrayList();
      }
      _inputStreams.add(stream);
      return stream;
   }

   /**
    * @return The name of the data.
    */
   public String getName()
   {
      return _name;
   }

   /**
    * @return An output stream that permits the client to read the byte array data.
    */
   public OutputStream getOutputStream() throws IOException
   {
      if (_data == null)
      {
         LOG.warn("Asked for an outputStream when byte array is null.");
         return null;
      }
      ByteArrayOutputStream stream = new ByteArrayOutputStream(_data.length + 100);
      stream.write(_data, 0, _data.length);
      return new ByteArrayDSOutputStream(stream, this);
   }
   
   private synchronized void removeStream(ByteArrayDSInputStream stream)
   {
      int size = _inputStreams != null ? _inputStreams.size() : 0;
      for (int i = 0; _inputStreams != null && i < size; i++)
      {
         if (_inputStreams.get(i) == stream)
         {
            _inputStreams.remove(i);
            break;
         }
      }
   }
   
   private synchronized void alertModified()
   {
      Iterator iter = _inputStreams.iterator();
      while (iter.hasNext())
      {
         ((ByteArrayDSInputStream) iter.next()).alertModified();
      }
   }

   private synchronized void setData(byte[] data)
   {
      _data = data;
   }
   
   private static final class ByteArrayDSInputStream
      extends InputStream
   {
      private ByteArrayInputStream _stream = null;
      private ByteArrayDataSource _owner = null;
      private boolean _isModified = false;
      
      private ByteArrayDSInputStream(ByteArrayInputStream stream, ByteArrayDataSource owner)
      {
         _stream = stream;
         _owner = owner;
      }
      
      public int available()
      {
         return _stream.available();
      }
      
      public void close()
         throws IOException
      {
         _stream.close();
         _owner.removeStream(this);
      }
      
      public void mark(int readAheadLimit)
      {
         _stream.mark(readAheadLimit);
      }
      
      public boolean markSupported()
      {
         return _stream.markSupported();
      }
      
      private void alertModified()
      {
         _isModified = true;
      }
      
      public int read()
      {
         if (_isModified)
         {
            throw new ConcurrentModificationException("The byte array in this data source has been written to at the same time as it is being read.");
         }
         return _stream.read();
      }
      
      public int read(byte[]b, int off, int len)
      {
         if (_isModified)
         {
            throw new ConcurrentModificationException("The byte array in this data source has been written to at the same time as it is being read.");
         }
         return _stream.read(b, off, len);
      }
      
      public void reset()
      {
         _stream.reset();
      }
      
      public long skip(long n)
      {
         return _stream.skip(n);
      }
   }
   
   private static final class ByteArrayDSOutputStream
      extends OutputStream
   {
      ByteArrayOutputStream _stream = null;
      ByteArrayDataSource _owner = null;
      
      public ByteArrayDSOutputStream(ByteArrayOutputStream stream, ByteArrayDataSource owner)
      {
         _stream = stream;
         _owner = owner;
      }
      
      public void flush()
         throws IOException
      {
         _stream.flush();
         _owner.setData(_stream.toByteArray());
      }
      
      public void close()
         throws IOException
      {
         _stream.close();
         _owner.setData(_stream.toByteArray());
      }
      
      public void write(byte[] b)
         throws IOException
      {
         _owner.alertModified();
         _stream.write(b);
      }
      
      public void write(byte[] b, int off, int len)
      {
         _stream.write(b, off, len);
      }
      
      public void write(int b)
      {
         _stream.write(b);
      }
   }
}
